/*--------------------------------------------------------------------
 * pitrap.c
 * FILE: pitrap.c
 * DESCRIPTION:
 *             This program calculates the value of PI using the
 *             "trapezoidal rule".  All processes contribute to
 *             the calculation of pi.  This version uses MPI_Gather
 *             to gather up individual integrals of each process.
 *             The master process (process 0) sums the integrals to
 *             give the value of PI.  The master process broadcasts
 *             the value of PI back to all the processes using 
 *             MPI_Bcast.
 * Author: Sameet Kudyadi, PHY-403, (315)-443-5981
 * Version: 1.0
 * Date: 9/28/98
 ---------------------------------------------------------------------*/ 

#include <stdio.h>
#include <mpi.h>
#include <math.h>
main(int argc, char* argv[])
{
        int my_rank;     /*  rank of my process  */
        int p;           /*  The number of processes  */
        double a=0.0;    /*  Left endpoint  */
        double b=1.0;    /*  Right endpoint  */
        double n=1024;   /*  Number of trapezoids used to calculate the integral  */
        double h;        /*  Trapezoid base length  */
        double local_a;  /*  Left endpoint of my process  */
        double local_b;  /*  Right endpoint of my process  */
        int local_n;     /*  Number of trapezoids for my calculation  */
        double integral; /*  Integral calculated by each process  */
        double total[10];/*  buffer containing each sub-integral  */
	double pi;       /*  the result of the computation of pi  */
        int j;           /*  used for looping  */
	
        double Trapezoid(double local_a, double local_b, int local_n, double h);	/*  calculate the local integral  */
	
        /*  Initializes MPI environment using command line arguments  */
        MPI_Init(&argc, &argv); 

        /*  Finds out how many processes are being used  */
        MPI_Comm_size(MPI_COMM_WORLD, &p);  

        /*  Get my process rank  */
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);  
	
        h = (b-a)/n;	/*  h is the same for all processes  */
	local_n = n/p;	/*  number of trapezoids for each process */
	local_a = a + my_rank*local_n*h;
	local_b = local_a + local_n*h;
	integral = Trapezoid(local_a, local_b, local_n, h);	
        
        /*  Each process sends it's integral to master process 0  */
        /*  Process 0 stores the integrals in rank order in array total  */
        MPI_Gather(&integral, 1, MPI_DOUBLE, total, 1, MPI_DOUBLE,0, MPI_COMM_WORLD);
        
        /*  Master process performs the summation to find PI  */
        /*  It broadcasts pi to all other processes  */
        /*  It displays the value of pi  */
        if (my_rank ==0) {
		for (j=0; j<p; j++)                           
                        pi = pi + total[j];		            
                MPI_Bcast(&pi, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
                printf("The value of pi is %.16f\n", pi);
	} 
        
        /*  Shuts down MPI  */
	MPI_Finalize;
}	/*  end of main function  */


/*---------<  Trapezoid is used to calculate the local integral  >--------*/
double Trapezoid(double local_a, double local_b, int local_n, double h)
{
	double integ;	/*  used to store the integral calculated by each process  */
	double x;
	int j;
	
	double f(double x);	/*  function that we are integrating over  */

	integ = (f(local_a) + f(local_b))/2.0;
	x = local_a;
	for (j = 1; j <= local_n-1; j++)
	{
		x= x + h;
		integ = integ + f(x);
	}
	integ = integ*h;
	return integ;
}

/*---------<   Calculates the given function  >-----------*/
double f(double x)
{
	double return_val;
	/*  Calculating the function 4/(1 + x*x)  */
	return_val = 4 /( 1 + x*x );
	return return_val;
}	
	
